"""pytest test cases for examples in fenced code blocks in README.md."""

import inspect
import re
import textwrap

import pytest

import phmdoctest.main
import phmdoctest.simulator
import phmdoctest.print_capture
import phmdoctest.tool
import verify

# Caution:
# This test file is run by pytest.
# The call to invoke_and_pytest() will start pytest in a
# subprocess.
# Pytest captures stdout and so does CliRunner.invoke().


readme_blocks = []


def setup_module():
    """Collect Markdown fenced code blocks contents from README.md.

    The test cases here must be run in order because they
    pop items from the list readme_blocks.

    This means using a pytest -k KEY more specific than
    "-k test_readme" risks taking the wrong block from the
    iterator readme_blocks causing all subsequent test_readme
    test case to fail.
    """
    # test cases below iterate through the blocks.
    global readme_blocks
    readme_blocks = iter(phmdoctest.tool.fenced_code_blocks('README.md'))


def test_raw_example1_md():
    """README raw markdown is same as file doc/example1.md."""
    want = next(readme_blocks)
    with open('doc/example1.md') as fp:
        got = fp.read()
        verify.a_and_b_are_the_same(want, got)


def test_example1():
    """Make sure generated --outfile is as expected; Run pytest.

    Check the copy of test_example1.py in the fenced code block.
    """
    # The helper checks the generated --outfile against the disk file.
    example1_command = next(readme_blocks)
    want = next(readme_blocks)
    _ = verify.one_example(
        example1_command,
        want_file_name='doc/test_example1.py',
        pytest_options=None
    )
    # Make sure the copy of test_example1.py in README.md
    # is the same as the disk file.
    with open('doc/test_example1.py') as fp:
        got = fp.read()
        verify.a_and_b_are_the_same(want, got)

    # Run again and call pytest to make sure the file works with pytest.
    simulator_status = verify.one_example(
        example1_command,
        want_file_name=None,
        pytest_options=['--strict', '-v']
    )
    assert simulator_status.pytest_exit_code == 0


def test_report():
    """README report output is same as produced by the command."""
    report_command = next(readme_blocks)
    want = next(readme_blocks)
    simulator_status = phmdoctest.simulator.run_and_pytest(
        report_command, pytest_options=None)
    got = simulator_status.runner_status.stdout
    verify.a_and_b_are_the_same(want, got)


def test_skip_example():
    """Make sure generated --outfile and --report are as expected."""
    skip_command = next(readme_blocks)
    want = next(readme_blocks)    # get the skip report
    short_form_command = next(readme_blocks)
    simulator_status = verify.one_example(
        skip_command,
        want_file_name='doc/test_example2.py',
        pytest_options=None
    )
    got1 = simulator_status.runner_status.stdout
    verify.a_and_b_are_the_same(want, got1)

    # test the first -s form of the --skip
    simulator_status = verify.one_example(
        short_form_command,
        want_file_name='doc/test_example2.py',
        pytest_options=None
    )
    got2 = simulator_status.runner_status.stdout
    verify.a_and_b_are_the_same(want, got2)


def test_outfile_to_stdout():
    """Make sure generated --outfile and --report are as expected."""
    outfile_command1 = next(readme_blocks)
    outfile_command2 = next(readme_blocks)
    simulator_status = verify.one_example(
        outfile_command1,
        want_file_name=None,
        pytest_options=None
    )
    with open('doc/test_example2.py') as fp:
        want = fp.read()
    got1 = simulator_status.runner_status.stdout
    verify.a_and_b_are_the_same(want, got1)

    simulator_status = verify.one_example(
        outfile_command2,
        want_file_name=None,
        pytest_options=None
    )
    got2 = simulator_status.runner_status.stdout
    verify.a_and_b_are_the_same(want, got2)


def test_usage():
    """Example usage near the bottom."""
    # Note- There may be differences in whitespace between
    #       the README.md fenced code block and the
    #       phmdoctest --help output caused by the
    #       the terminal width when the installed
    #       phmdoctest command is executed.
    # Note- This test runs phmdoctest by calling the simulator
    #       which calls  Click.CliRunner.invoke().  invoke()
    #       displays entry-point as the calling program.
    #       The test here replaces 'entry-point' with 'phmdoctest'.
    simulator_status = phmdoctest.simulator.run_and_pytest(
        'phmdoctest --help', pytest_options=None)
    want1 = next(readme_blocks)
    want2 = re.sub(r'\s+', ' ', want1)

    got1 = simulator_status.runner_status.stdout
    got2 = got1.replace('entry-point', 'phmdoctest', 1)
    got3 = re.sub(r'\s+', ' ', got2)
    verify.a_and_b_are_the_same(want2, got3)


def test_yaml():
    """Show Markdown example and .travis.yml have the same commands."""
    markdown_example_text = next(readme_blocks)
    expected = """\
dist: xenial
language: python
sudo: false

matrix:
  include:
    - python: 3.5
      install:
        - pip install "." pytest
      script:
        - mkdir tests/tmp
        - phmdoctest project.md --report --outfile tests/tmp/test_project_readme.py
        - pytest --strict -vv tests"""
    verify.a_and_b_are_the_same(expected, markdown_example_text)
    with open('.travis.yml', 'r', encoding='utf-8') as f:
        travis_text = f.read()
        assert travis_text.startswith(expected)


# Developers: Changes here must be mirrored in a fenced code block in README.md.
# Runnable version of example code in README.md.
# The guts of this function are an exact copy of example in README.md.
def example_code():
    import phmdoctest.simulator
    command = 'phmdoctest doc/example1.md --report --outfile test_me.py'
    simulator_status = phmdoctest.simulator.run_and_pytest(
        well_formed_command=command,
        pytest_options=['--strict', '-v']
    )
    assert simulator_status.runner_status.exit_code == 0
    assert simulator_status.pytest_exit_code == 0


def test_simulator_python_code():
    """Assure the guts of example_code() above are same as in Markdown."""
    want1 = next(readme_blocks)
    want2 = textwrap.indent(want1, '    ')
    got1 = inspect.getsource(example_code)
    got2 = got1.replace('def example_code():\n', '')
    verify.a_and_b_are_the_same(want2, got2)
    # also make sure the code runs with no assertions
    example_code()


def test_consumed_all_fenced_code_blocks():
    """Verify no left over fenced code blocks from README.md"""
    with pytest.raises(StopIteration):
        _ = next(readme_blocks)



"""Simulate running a phmdoctest command. Used for testing.

Use during development to verify phmdoctest console commands
will succeed when called by Continuous Integration scripts.
If --outfile writes a file, the file is written in a
temporary directory.
Optionally run pytest on the temporary file.
"""

import os.path
import re
import subprocess
from tempfile import TemporaryDirectory
from typing import List, Optional, NamedTuple

import click.testing

from .main import entry_point


SimulatorStatus = NamedTuple(
    'SimulatorStatus',
    [('runner_status', click.testing.Result),
     ('outfile', Optional[str]),
     ('pytest_exit_code', Optional[int])
     ])
"""run_and_pytest() return value."""


def run_and_pytest(
        well_formed_command: str,
        pytest_options: Optional[List[str]] = None) -> SimulatorStatus:
    """
    Simulate a phmdoctest command, optionally run pytest.

    If a filename is provided by the --outfile option, the
    command is rewritten replacing the OUTFILE with a
    path to a temporary directory and a synthesized filename.

    To run pytest on an --outfile, pass a list of zero or
    more pytest_options.  pytest is run in a subprocess.

    The PYPI package pytest must be installed separately
    since pytest is not required to install phmdoctest.
    Use this command: ``pip install pytest``

    Returns SimulatorStatus object.
    SimulatorStatus.runner_status is the CliRunner.invoke return value.

    If an outfile is written or streamed to stdout a copy of it
    is found in simulator_status.runner_status.stdout.

    If calling run_and_pytest() from a pytest file, try adding the
    pytest option ``--capture=tee-sys`` to the command running
    pytest on the file.

    For example on a checkout of phmdoctest the command line:

    ``python -m pytest tests -v --capture=tee-sys``

    will print the outputs from the subprocess.run() invocations
    of pytest on the --outfile written to the temporary directory.
    A wild guess would be that the subprocess inherited changes
    made to the parent by --capture=tee-sys.

    Args:
        well_formed_command
            - starts with phmdoctest
            - followed by MARKDOWN_FILE
            - ends with --outfile OUTFILE (if needed)
            - all other options are between MARKDOWN_FILE and --outfile
              for example:
              ``phmdoctest MARKDOWN_FILE --skip FIRST --outfile OUTFILE``

        pytest_options
            List of strings like this: ``['--strict', '-v']``.
            Set to empty list to run pytest with no options.
            Set to None to skip pytest.

    Returns:
        SimulatorStatus containing runner_status, outfile,
        and pytest_exit_code.
    """
    assert well_formed_command.startswith('phmdoctest ')
    # trim off any trailing whitespace
    command0 = well_formed_command.rstrip()
    # chop off phmdoctest since invoking by a python function call
    command1 = command0.replace('phmdoctest ', '', 1)
    # simulate commands that don't write OUTFILE.
    wants_help = '--help' in command1
    wants_version = '--version' in command1
    stream_outfile = (
            command1.endswith('--outfile -') or
            command1.endswith('--outfile=-')
    )
    no_outfile = '--outfile' not in command1
    runner = click.testing.CliRunner()
    if wants_help or wants_version or stream_outfile or no_outfile:
        return SimulatorStatus(
            runner_status=runner.invoke(cli=entry_point, args=command1),
            outfile=None,
            pytest_exit_code=None
        )

    # Simulate commands that write an OUTFILE.
    # Split up the command into pieces.
    # Chop out the path to the markdown file.
    # Drop the rest of the command starting at --outfile and the
    # outfile path since we rename the outfile in the invoked command.
    with TemporaryDirectory() as tmpdirname:
        # Create a filename in the temporary directly to
        # receive the OUTFILE.
        # Rewrite the command to use the new OUTFILE path and
        # split up the command to a list of strings.
        # Calling invoke with the single string form of the
        # rewritten command fails to find the outfile.
        # This might be because it is now an absolute path
        # to the tmpdir.
        outfile_path = os.path.join(tmpdirname, 'test_sim_tmpdir.py')
        markdown_path, command2 = command1.split(maxsplit=1)
        command3 = command2[:command2.find('--outfile')].strip()
        phm_args = [markdown_path]

        # Split up the rest of the command into pieces to pass to
        # runner.invoke().
        #
        # Developers:
        # Since the --outfile part has already been removed from command3
        # the only possible option remaining that takes TEXT is --skip.
        # If a new option that takes TEXT is added, add code here
        # to replace its '='.
        #
        # Special code to handle a --skip TEXT where TEXT is double quoted.
        # For example
        #    --skip="Python 3.7"
        #         or
        #    --skip "Python 3.7"
        command4 = command3.replace('--skip=', '--skip ')
        # get characters between double quotes including the quotes
        # get runs of non-whitespace characters
        args1 = re.findall(pattern=r'("[^"]*"|\S+)', string=command4)
        # If both leading and trailing double quotes, remove them.
        args2 = [re.sub('^"([^"]*)"$', r'\1', arg) for arg in args1]
        phm_args.extend(args2)
        phm_args.extend(['--outfile', outfile_path])
        runner_status = runner.invoke(cli=entry_point, args=phm_args)

        # return now if the command failed
        if runner_status.exit_code:
            return SimulatorStatus(
                runner_status=runner_status,
                outfile=None,
                pytest_exit_code=None
            )

        # Copy the generated pytest file from the isolated filesystem.
        with open(outfile_path, 'r', encoding='utf-8') as fp:
            outfile_text = fp.read()

        pytest_exit_code = None
        if pytest_options is not None:
            completed = subprocess.run(
                ['python', '-m', 'pytest'] + pytest_options + [tmpdirname],
            )
            pytest_exit_code = completed.returncode

        return SimulatorStatus(
            runner_status=runner_status,
            outfile=outfile_text,
            pytest_exit_code=pytest_exit_code
        )



"""General purpose tools get fenced code blocks from Markdown."""

from typing import IO, List

import commonmark    # type: ignore
import commonmark.node    # type: ignore


def fenced_code_blocks(markdown_filename: str) -> List[str]:
    """Return Markdown fenced code block contents as a list of strings.

    Args:
        markdown_filename
            Path to the Markdown file.

    Returns:
        List of strings, one for the contents of each Markdown
        fenced code block.
    """
    with open(markdown_filename, encoding='utf-8') as fp:
        nodes = fenced_block_nodes(fp)
        return [node.literal for node in nodes]


def fenced_block_nodes(fp: IO[str]) -> List[commonmark.node.Node]:
    """Get markdown fenced code blocks as list of Node objects.

    Args:
        fp
            file object returned by open().

    Returns:
         List of commonmark.node.Node objects.
    """
    nodes = []
    doc = fp.read()
    parser = commonmark.Parser()
    ast = parser.parse(doc)
    walker = ast.walker()
    # Presumably, because fenced code blocks nodes are leaf nodes
    # they will only be entered once by the walker.
    for node, entering in walker:
        if node.t == 'code_block' and node.is_fenced:
            nodes.append(node)
    return nodes



"""pytest test case file code generator."""

import inspect
from itertools import zip_longest
import textwrap


def line_by_line_compare_exact(a, b):
    """Line by line helper compare function with assertion for pytest."""
    a_lines = a.splitlines()
    b_lines = b.splitlines()
    for a_line, b_line in zip_longest(a_lines, b_lines):
        assert a_line == b_line


# The function below is used as a template to generate python source
# code to be written to a file.
# It is coded here as compiled python so the IDE can check for
# syntax and style.
# Python introspection of the function's source code provides the
# source code as a string.
#
# This template will be customized by replacing:
# - The _identifier substring of the function name.
# - Insert example code indented 4 spaces on a new line
#   after the def statement.
# - Triple quoted string contents with the expected output.
def test_identifier(capsys):
    expected_str = """\
<<<replaced>>>"""
    line_by_line_compare_exact(a=expected_str, b=capsys.readouterr().out)


class PytestFile:
    def __init__(self, description: str = ''):
        docstring = '"""' + description + '"""'
        self.lines = [docstring]
        self.lines.append('from itertools import zip_longest')
        self._empty_line()
        self._empty_line()
        # copy the helper function def
        self.lines.append(inspect.getsource(line_by_line_compare_exact))

    def __str__(self) -> str:
        return '\n'.join(self.lines)

    def _empty_line(self) -> None:
        """Add an empty line to the file."""
        self.lines.append('')

    @staticmethod
    def _remove_output_check(source: str) -> str:
        """Replace the expected output with a Caution message."""
        ix = source.index('    expected_str = """')
        source = source[:ix] + '    # Caution- no assertions.\n'
        return source

    def add_test_case(
            self, identifier: str, code: str, expected_output: str) -> None:
        """Add a def test_ function with code and comparison logic."""
        assert identifier.isidentifier(), 'must be a valid python identifier'
        self._empty_line()
        src = inspect.getsource(test_identifier)
        src = src.replace('identifier', identifier, 1)

        # indent contents of code block and place after '(capysy):\n'
        indented_code = textwrap.indent(code, '    ')
        src = src.replace('(capsys):', '(capsys):\n' + indented_code, 1)

        if expected_output:
            src = src.replace('<<<replaced>>>', expected_output, 1)
        else:
            src = self._remove_output_check(src)
        self.lines.append(src)

    def add_source(self, source: str) -> None:
        """Add the source code as is to the generated test file."""
        self._empty_line()
        self.lines.append(source)



"""pytest file built from doc/example2.md"""
from itertools import zip_longest


def line_by_line_compare_exact(a, b):
    """Line by line helper compare function with assertion for pytest."""
    a_lines = a.splitlines()
    b_lines = b.splitlines()
    for a_line, b_line in zip_longest(a_lines, b_lines):
        assert a_line == b_line


def test_code_9_output_14(capsys):
    squares = [1, 4, 9, 16, 25]
    print(squares)

    expected_str = """\
[1, 4, 9, 16, 25]
"""
    line_by_line_compare_exact(a=expected_str, b=capsys.readouterr().out)


def test_code_37(capsys):
    a, b = 0, 1
    while a < 1000:
        print(a, end=',')
        a, b = b, a+b

    # Caution- no assertions.


def test_code_44_output_51(capsys):
    words = ['cat', 'window', 'defenestrate']
    for w in words:
        print(w, len(w))

    expected_str = """\
cat 3
window 6
defenestrate 12
"""
    line_by_line_compare_exact(a=expected_str, b=capsys.readouterr().out)



"""Check test_example2.md includes same code as doc/test_example2.py."""

import phmdoctest.tool
import verify


def test_markdown_wrapper_test_example2():
    """Fenced code block in test_example2.md == file test_example2.py."""
    with open('doc/test_example2.py', 'r', encoding='utf-8') as f:
        want = f.read()
    blocks = phmdoctest.tool.fenced_code_blocks('doc/test_example2.md')
    assert len(blocks) == 1
    got = blocks[0]
    verify.a_and_b_are_the_same(a=want, b=got)



"""pytest test cases for phmdoctest."""

import re

import click
import pytest

import phmdoctest
import phmdoctest.main
import phmdoctest.simulator
import phmdoctest.print_capture
import verify


# Caution:
# This test file is run by pytest.
# The call to invoke_and_pytest() will start pytest in a
# subprocess.
# Pytest captures stdout and so does CliRunner.invoke().

class TestSameVersions:
    """Verify same release version string in all places.

    Obtain the version string from various places in the source tree
    and check that they are all the same.
    Compare all the occurrences to phmdoctest.__version__.
    This test does not prove the version is correct.
    Whitespace may be significant in some cases.
    """
    package_version = phmdoctest.__version__

    def verify_found_in_file(self, filename, format_spec='{}'):
        """Format the package version and look for result in caller's file."""
        looking_for = format_spec.format(self.package_version)
        with open(filename, 'r', encoding='utf-8') as f:
            text = f.read()
        assert looking_for in text

    def test_readme_md(self):
        """Check the version near the top of README.md."""
        self.verify_found_in_file('README.md', '# phmdoctest {}')

    def test_index_rst(self):
        """Check the version is anywhere in index.rst."""
        self.verify_found_in_file('index.rst', 'phmdoctest {}\n=============')

    def test_recent_changes(self):
        """Check the version is anywhere in recent_changes.md."""
        self.verify_found_in_file('doc/recent_changes.md', '{} - ')

    def test_conf_py_release(self):
        """Check version in the release = line in conf.py."""
        self.verify_found_in_file('conf.py', "release = '{}'")

    def test_setup_py(self):
        """Check the version anywhere in setup.py."""
        with open('setup.py', 'r', encoding='utf-8') as f:
            setup_text = f.read()
        # keep the part between single or double quotes after version=
        match = re.search(r" *version=['\"]([^'\"]*)['\"]", setup_text, re.M)
        assert match.group(1) == self.package_version


class TestDocBuildVersions:
    """
    Some versions are the same in doc/requirements.txt and setup.py.

    Click and monotable versions should be the same.

    For the Sphinx documentation build on readthedocs.org (RTD)
    specific versions are pinned by the file doc/requirements.txt.

    For Sphinx autodoc the phmdoctest dependencies Click and monotable
    are installed so that the RTD build can import phmdoctest to look
    for docstrings.

    Note that commonmark is also a phmdoctest dependency. Because it is
    pinned to different versions in doc/requirements.txt and setup.py
    it is not tested here.
    """
    with open('doc/requirements.txt', 'r', encoding='utf-8') as f:
        doc_requirements = f.read()
        for line in doc_requirements.splitlines():
            if line.startswith('Click'):
                click_version = line
            if line.startswith('monotable'):
                monotable_version = line
    with open('setup.py', 'r', encoding='utf-8') as f:
        setup = f.read()

    @staticmethod
    def to_setup_style(value):
        """Convert value from requirements.txt style to setup.py style."""
        drop_newline = value.replace('\n', '', 1)
        drop_space = drop_newline.replace(' ', '', 1)
        quoted = drop_space.join(["'", "'"])
        return quoted

    def test_click(self):
        """Click version in doc/requirements.txt same as setup.py"""
        assert self.click_version in self.doc_requirements, 'sanity check'
        expected = self.to_setup_style(self.click_version)
        assert self.setup.count(expected) == 1

    def test_monotable(self):
        """monotable version in doc/requirements.txt same as setup.py"""
        assert self.monotable_version in self.doc_requirements, 'sanity check'
        expected = self.to_setup_style(self.monotable_version)
        assert self.setup.count(expected) == 1


def test_def_test_nothing_fails():
    """This is done for code coverage of the function."""
    with pytest.raises(AssertionError):
        phmdoctest.main.test_nothing_fails()


def test_def_test_nothing_passes():
    """This is done for code coverage of the function."""
    phmdoctest.main.test_nothing_passes()


def test_empty_output_block_fails():
    """Empty output block causes phmdoctest to raise AssertionError."""
    command = (
        'phmdoctest tests/empty_output_block.md --outfile discarded.py'
    )
    simulator_status = phmdoctest.simulator.run_and_pytest(
        well_formed_command=command,
        pytest_options=None
    )
    exc = simulator_status.runner_status.exception
    assert 'zero length expected output block' in str(exc)
    assert simulator_status.runner_status.exit_code == 1


def test_code_does_not_print_fails():
    """Show empty stdout mis-compares with non-empty output block."""
    command = (
        'phmdoctest tests/does_not_print.md --outfile discarded.py'
    )
    simulator_status = phmdoctest.simulator.run_and_pytest(
        well_formed_command=command,
        pytest_options=['--strict', '-v']
    )
    assert simulator_status.runner_status.exit_code == 0
    assert simulator_status.pytest_exit_code == 1


def test_more_printed_than_expected_fails():
    """Show pytest fails when more lines are printed than expected."""
    command = (
        'phmdoctest tests/missing_some_output.md --outfile discarded.py'
    )
    simulator_status = phmdoctest.simulator.run_and_pytest(
        well_formed_command=command,
        pytest_options=['--strict', '-v']
    )
    assert simulator_status.runner_status.exit_code == 0
    assert simulator_status.pytest_exit_code == 1


def test_more_expected_than_printed_fails():
    """Show pytest fails when more lines are printed than expected."""
    command = (
        'phmdoctest tests/extra_line_in_output.md --outfile discarded.py'
    )
    simulator_status = phmdoctest.simulator.run_and_pytest(
        well_formed_command=command,
        pytest_options=['--strict', '-v']
    )
    assert simulator_status.runner_status.exit_code == 0
    assert simulator_status.pytest_exit_code == 1


def test_skip_same_block_twice():
    """Show identifying a skipped code block more than one time is OK."""
    command = (
        'phmdoctest doc/example2.md --skip "Python 3.7" --skip LAST'
        ' --skip LAST --report --outfile discarded.py'
    )
    simulator_status = phmdoctest.simulator.run_and_pytest(
        well_formed_command=command,
        pytest_options=['--strict', '-v']
    )
    assert simulator_status.runner_status.exit_code == 0
    assert simulator_status.pytest_exit_code == 0


def test_pytest_really_fails():
    """Make sure pytest fails due to incorrect expected output in the .md.

    Generate a pytest that will assert.
    """
    simulator_status = verify.one_example(
        'phmdoctest tests/unexpected_output.md --outfile test_unexpected_output.py',
        want_file_name=None,
        pytest_options=['--strict', '-v']
    )
    assert simulator_status.pytest_exit_code == 1


def test_def_test_identifier():
    """Painful way to eliminate 2 coverage missed statements."""
    # The function print_capture.test_identifier() is used as
    # a template to generate Python code.
    # It accepts the pytest fixture called capsys when the
    # generated pytest is run.
    # phmodctest doesn't call this function so it shows up
    # in the coverage report as a missed statement.
    # Here a test mock up of the fixture is created that
    # provides the expected value as its out attribute.
    class MockReadouterr:
        def __init__(self):
            self.out = '<<<replaced>>>'

    class MockCapsys:
        @staticmethod
        def readouterr():
            return MockReadouterr()

    phmdoctest.print_capture.test_identifier(MockCapsys())


def test_blanklines_in_output():
    """Expected output has empty lines and does not have doctest <BLANKLINE>."""
    command = (
        'phmdoctest tests/output_has_blank_lines.md --outfile discarded.py'
    )
    simulator_status = phmdoctest.simulator.run_and_pytest(
        well_formed_command=command,
        pytest_options=['--strict', '-v']
    )
    assert simulator_status.runner_status.exit_code == 0
    assert simulator_status.pytest_exit_code == 0


def test_skip_first():
    """Verify --skip FIRST."""
    command = (
        'phmdoctest doc/example2.md --skip "Python 3.7" -sFIRST'
        ' --skip LAST --report --outfile discarded.py'
    )
    simulator_status = phmdoctest.simulator.run_and_pytest(
        well_formed_command=command,
        pytest_options=['--strict', '-v']
    )
    assert simulator_status.runner_status.exit_code == 0
    assert simulator_status.pytest_exit_code == 0
    stdout = simulator_status.runner_status.stdout
    assert 'py3         9  skip-code    "FIRST"' in stdout
    assert 'FIRST         9' in stdout


def test_skip_second():
    """Verify --skip SECOND."""
    command = (
        'phmdoctest doc/example2.md --skip SECOND'
        ' --report --outfile discarded.py'
    )
    simulator_status = phmdoctest.simulator.run_and_pytest(
        well_formed_command=command,
        pytest_options=['--strict', '-v']
    )
    assert simulator_status.runner_status.exit_code == 0
    assert simulator_status.pytest_exit_code == 0
    stdout = simulator_status.runner_status.stdout
    assert 'py3        20  skip-code    "SECOND"' in stdout
    assert 'SECOND        20' in stdout


def test_skip_second_when_only_one():
    """Verify --skip SECOND selects no block when only 1 code block."""
    command = (
        'phmdoctest doc/example1.md -sFIRST'
        ' --skip SECOND --report --outfile discarded.py'
    )
    simulator_status = phmdoctest.simulator.run_and_pytest(
        well_formed_command=command,
        pytest_options=['--strict', '-v']
    )
    assert simulator_status.runner_status.exit_code == 0
    assert simulator_status.pytest_exit_code == 0
    assert 'def test_nothing_passes()' in simulator_status.outfile
    assert 'SECOND\n' in simulator_status.runner_status.stdout


def test_skip_second_when_more_than_one():
    """Verify --skip SECOND when more than 1 code block."""
    command = (
        'phmdoctest doc/example2.md -sFIRST'
        ' --skip SECOND --report --outfile discarded.py'
    )
    simulator_status = phmdoctest.simulator.run_and_pytest(
        well_formed_command=command,
        pytest_options=['--strict', '-v']
    )
    assert simulator_status.runner_status.exit_code == 0
    assert simulator_status.pytest_exit_code == 0
    stdout = simulator_status.runner_status.stdout
    assert 'py3        20  skip-code    "SECOND"' in stdout
    assert 'SECOND        20' in stdout


def test_skip_code_that_has_no_output_block():
    """Skip code with no output block."""
    command = (
        'phmdoctest doc/example2.md --skip SECOND --skip="while a < 1000:"'
        ' --report --outfile discarded.py'
    )
    simulator_status = phmdoctest.simulator.run_and_pytest(
        well_formed_command=command,
        pytest_options=['--strict', '-v']
    )
    assert simulator_status.runner_status.exit_code == 0
    assert simulator_status.pytest_exit_code == 0
    stdout = simulator_status.runner_status.stdout
    assert 'py3        20  skip-code    "SECOND"' in stdout
    assert 'py3        37  skip-code    "while a < 1000:"' in stdout
    assert 'SECOND           20' in stdout
    assert 'while a < 1000:  37' in stdout


def test_skip_matches_start_of_contents():
    """Skip pattern matching first characters of code block."""
    command = (
        'phmdoctest doc/example2.md --skip SECOND --skip="Python 3.7"'
        '  --skip="words =" --report --outfile discarded.py'
    )
    simulator_status = phmdoctest.simulator.run_and_pytest(
        well_formed_command=command,
        pytest_options=['--strict', '-vv']
    )
    assert simulator_status.runner_status.exit_code == 0
    assert simulator_status.pytest_exit_code == 0
    stdout = simulator_status.runner_status.stdout
    assert 'words =       44' in stdout


# words = ['cat', 'window', 'defenestrate']

def test_multiple_skips_report():
    """More than one skip applied to the same Python code block."""
    command = 'phmdoctest doc/example2.md --report -sprint -slen'
    simulator_status = phmdoctest.simulator.run_and_pytest(
        well_formed_command=command,
        pytest_options=None
    )
    assert simulator_status.runner_status.exit_code == 0
    stdout = simulator_status.runner_status.stdout
    assert '                            "len"' in stdout
    assert 'len           44' in stdout


def test_no_blocks_left_to_test_passing():
    """Generate a pytest file that passes when no blocks to test."""
    command = (
        'phmdoctest doc/example1.md -sFIRST'
        ' --skip SECOND --report --outfile discarded.py'
    )
    simulator_status = phmdoctest.simulator.run_and_pytest(
        well_formed_command=command,
        pytest_options=['--strict', '-v']
    )
    assert simulator_status.runner_status.exit_code == 0
    assert simulator_status.pytest_exit_code == 0
    assert 'def test_nothing_passes()' in simulator_status.outfile


def test_no_blocks_left_to_test_fails():
    """Generate a pytest file that asserts when no blocks to test."""
    command = (
        'phmdoctest doc/example1.md -sFIRST --fail-nocode'
        ' --skip SECOND --report --outfile discarded.py'
    )
    simulator_status = phmdoctest.simulator.run_and_pytest(
        well_formed_command=command,
        pytest_options=['--strict', '-v']
    )
    assert simulator_status.runner_status.exit_code == 0
    assert simulator_status.pytest_exit_code == 1    # pytest failed
    assert 'def test_nothing_fails()' in simulator_status.outfile


def test_no_code_blocks():
    """Process .md that has no code blocks."""
    command = (
        'phmdoctest tests/no_code_blocks.md'
        ' --report --outfile discarded.py'
    )
    simulator_status = phmdoctest.simulator.run_and_pytest(
        well_formed_command=command,
        pytest_options=['--strict', '-v']
    )
    assert simulator_status.runner_status.exit_code == 0
    assert simulator_status.pytest_exit_code == 0
    assert 'def test_nothing_passes()' in simulator_status.outfile


def test_missing_markdown_file():
    """Usage error for MARKDOWN_FILE that does not exist."""
    command = 'phmdoctest tests/bogus.md --outfile discarded.py'
    simulator_status = phmdoctest.simulator.run_and_pytest(
        well_formed_command=command,
        pytest_options=['--strict', '-v']
    )
    assert (
        simulator_status.runner_status.exit_code ==
        click.UsageError.exit_code
    )
    assert simulator_status.outfile is None
    assert simulator_status.pytest_exit_code is None


def test_bad_usage_option():
    """Usage error for misspelled option."""
    command = 'phmdoctest doc/example1.md --troper --outfile discarded.py'
    simulator_status = phmdoctest.simulator.run_and_pytest(
        well_formed_command=command,
        pytest_options=['--strict', '-v']
    )
    assert (
        simulator_status.runner_status.exit_code == click.UsageError.exit_code
    )
    assert simulator_status.outfile is None
    assert simulator_status.pytest_exit_code is None



"""Logic used by pytest test cases."""

from itertools import zip_longest

import phmdoctest.print_capture
import phmdoctest.simulator


def a_and_b_are_the_same(a, b):
    """Line by line helper compare function with assertion for pytest."""
    a_lines = a.splitlines()
    b_lines = b.splitlines()
    for a_line, b_line in zip_longest(a_lines, b_lines):
        assert a_line == b_line, a_line + '|' + b_line


def one_example(
        well_formed_command,
        want_file_name=None,
        pytest_options=None):
    """Simulate running a phmdoctest command and pytest on the result."""
    simulator_status = phmdoctest.simulator.run_and_pytest(
        well_formed_command, pytest_options=pytest_options)
    # check that the phmdoctest command succeeded
    assert simulator_status.runner_status.exit_code == 0

    # check the OUTFILE against the expected value
    if want_file_name is not None:
        with open(want_file_name) as f:
            want = f.read()
            a_and_b_are_the_same(simulator_status.outfile, want)
    return simulator_status



"""pytest file built from doc/example1.md"""
from itertools import zip_longest


def line_by_line_compare_exact(a, b):
    """Line by line helper compare function with assertion for pytest."""
    a_lines = a.splitlines()
    b_lines = b.splitlines()
    for a_line, b_line in zip_longest(a_lines, b_lines):
        assert a_line == b_line


def test_code_4_output_17(capsys):
    from enum import Enum

    class Floats(Enum):
        APPLES = 1
        CIDER = 2
        CHERRIES = 3
        ADUCK = 4
    for floater in Floats:
        print(floater)

    expected_str = """\
Floats.APPLES
Floats.CIDER
Floats.CHERRIES
Floats.ADUCK
"""
    line_by_line_compare_exact(a=expected_str, b=capsys.readouterr().out)



""" See The Python Standard Library, Runtime Services, __main__.

    This __main__.py module is included to execute code in
    the package when it is run with python -m phmdoctest.
"""

from .main import entry_point
entry_point()



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
py_sources_path = os.path.abspath('./src')
sys.path.insert(0, py_sources_path)

# -- Project information -----------------------------------------------------

# This file is placed in the project root directory rather than /doc.

# Configuration for Sphinx 1.8.5

project = 'phmdoctest'
copyright = '2020, Mark Taylor'
author = 'Mark Taylor'

# The full version, including alpha/beta/rc tags
release = '0.0.6'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ['recommonmark',
              'sphinx.ext.autodoc',
              'sphinx.ext.napoleon']

source_suffix = {
    '.rst': 'restructuredtext',
    '.md': 'markdown',
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# tmarktaylor: The documentation sources are at the project root.
# Any .md, .rst, or folders at the project root that don't
# belong in the documentation should be listed here.
#
# tmarktaylor: Travis CI Job #28.3 build log of tox.ini docs is finding
# input files in the .tox tree.

#
# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [
    'project.md', 'tests', 'src',
    '.tox',
    '.pytest_cache',
    '_build', 'Thumbs.db', '.DS_Store'
]

master_doc = 'index'

# -- Options for HTML output -------------------------------------------------

on_rtd = os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:
    import sphinx_rtd_theme
    html_theme = 'sphinx_rtd_theme'
    html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
# html_theme = 'alabaster'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = []  # ['_static']



from collections import namedtuple
from collections import Counter
from enum import Enum
import inspect
from typing import List, Optional

import click
import commonmark.node    # type: ignore
import monotable    # type: ignore

from . import tool
from . import print_capture


class Role(Enum):
    """Role that markdown fenced code block plays in testing."""
    UNKNOWN = '--'
    CODE = 'code'
    OUTPUT = 'output'
    SKIP_CODE = 'skip-code'
    SKIP_OUTPUT = 'skip-output'


class FencedBlock:
    """Selected fields from commonmark node plus new field role."""
    def __init__(self, node: commonmark.node.Node) -> None:
        """Extract fields from commonmark fenced code block node."""
        self.type = node.info
        self.line = node.sourcepos[0][0] + 1
        self.role = Role.UNKNOWN
        self.contents = node.literal
        self.output = None    # type: Optional['FencedBlock']
        self.skip_reasons = list()    # type: List[str]

    def set(self, role: Role) -> None:
        """Set the role for the fenced code block in subsequent testing."""
        self.role = role

    def set_link_to_output(self, fenced_block: 'FencedBlock') -> None:
        """Save a reference to the code block's output block."""
        assert self.role == Role.CODE, 'only allowed to be code'
        assert fenced_block.role == Role.OUTPUT, 'only allowed to be output'
        self.output = fenced_block

    def get_output(self) -> Optional['FencedBlock']:
        """Get this code block's associated output block."""
        return self.output

    def skip(self, reason: str) -> None:
        """Skip an already designated code block. Re-skip is OK."""
        assert self.role in (Role.CODE, Role.SKIP_CODE), (
            'Block at line {} cannot be skipped for pattern'
            ' {} since\nthe block is designated as {}'
            ' and must be either {} or {}.'
            ).format(
            self.line, reason, self.role.value,
            Role.CODE.value, Role.SKIP_CODE.value)
        if self.role == Role.CODE:
            self.set(Role.SKIP_CODE)
            output_block = self.get_output()
            if output_block:
                output_block.set(Role.SKIP_OUTPUT)
        self.skip_reasons.append(reason)


Args = namedtuple(
    'Args',
    [
        'markdown_file',
        'outfile',
        'skips',
        'is_report',
        'fail_nocode'
    ]
)
"""Command line arguments with some renames."""


@click.command()
@click.argument(
    'markdown_file',
    nargs=1,
    type=click.Path(
        exists=True,
        dir_okay=False,
        allow_dash=True,    # type: ignore
    )
)
@click.option(
    '--outfile',
    nargs=1,
    help=(
        'Write generated test case file to path TEXT. "-"'
        ' writes to stdout.'
    )
)
@click.option(
    '-s', '--skip',
    multiple=True,
    help=(
        'Any Python block that contains the substring TEXT is not tested.'
        ' More than one --skip TEXT is ok.'
        ' Double quote if TEXT contains spaces.'
        ' For example --skip="python 3.7" will skip every Python block that'
        ' contains the substring "python 3.7".'
        ' If TEXT is one of the 3 capitalized strings FIRST SECOND LAST'
        ' the first, second, or last Python block in the'
        ' Markdown file is skipped.'
        ' The fenced code block info string is not searched.'
    )
)
@click.option(
    '--report',
    is_flag=True,
    help='Show how the Markdown fenced code blocks are used.'
)
@click.option(
    '--fail-nocode',
    is_flag=True,
    help=(
        'This option sets behavior when the Markdown file has no Python'
        ' fenced code blocks or if all such blocks are skipped.'
        ' When this option is present the generated pytest file'
        ' has a test function called test_nothing_fails() that'
        ' will raise an assertion.'
        ' If this option is not present the generated pytest file'
        ' has test_nothing_passes() which will never fail.'
    )
)
@click.version_option()
# Note- docstring for entry point shows up in click's usage text.
def entry_point(markdown_file, outfile, skip, report, fail_nocode):
    args = Args(
        markdown_file=markdown_file,
        outfile=outfile,
        skips=skip,
        is_report=report,
        fail_nocode=fail_nocode,
    )

    # Find markdown blocks and pair up code and output blocks.
    with click.open_file(args.markdown_file, encoding='utf-8') as fp:
        blocks = convert_nodes(tool.fenced_block_nodes(fp))
    identify_code_and_output_blocks(blocks)

    code_blocks = [b for b in blocks if b.role == Role.CODE]

    apply_skips(args, code_blocks)

    if args.is_report:
        print_report(args, blocks)

    # build test cases and write to the --outfile path
    if args.outfile:
        test_case_string = build_test_cases(args, blocks)
        with click.open_file(args.outfile, 'w', encoding='utf-8') as ofp:
            ofp.write(test_case_string)


def convert_nodes(nodes: List[commonmark.node.Node]) -> List[FencedBlock]:
    """Create FencedBlock objects from commonmark fenced code block nodes."""
    blocks = []
    for node in nodes:
        blocks.append(FencedBlock(node))
    return blocks


PYTHON_FLAVORS = ['python', 'py3', 'python3']
"""Python fenced code blocks info string will start with one of these."""


def identify_code_and_output_blocks(blocks: List[FencedBlock]) -> None:
    """
    Designate which blocks are Python and guess which are output.

    The block.type is a copy of the Markdown fenced code block info_string.
    This string may start with the language intended for syntax coloring.
    A block is an output block if it has an empty markdown info field
    and follows a designated python code block.
    """

    for block in blocks:
        for flavor in PYTHON_FLAVORS:
            if block.type.startswith(flavor):
                block.set(Role.CODE)
                continue

    # When we find an output block we update the preceding
    # code block with a link to it.
    previous_block = None
    for block in blocks:
        if previous_block is not None:
            if not block.type and previous_block.role == Role.CODE:
                block.set(Role.OUTPUT)
                previous_block.set_link_to_output(block)
        previous_block = block
    # If we didn't find an output block for a code block
    # it can still be run, but there will be no comparison
    # to expected output.  If assertions are needed, they can
    # be added to the code block.


def apply_skips(args: Args, code_blocks: List[FencedBlock]) -> None:
    """Skip code blocks identified by patterns 'FIRST', 'SECOND', 'LAST'"""
    number_of_code_blocks = len(code_blocks)
    if number_of_code_blocks:
        for pattern in args.skips:
            index = None
            if pattern == 'FIRST':
                index = 0
            elif pattern == 'LAST':
                index = -1
            elif pattern == 'SECOND' and number_of_code_blocks > 1:
                index = 1
            if index is not None:
                code_blocks[index].skip(pattern)

    # skip code blocks identified by pattern matches with the code
    # Try to find each skip pattern in each code block.
    # If there is a match skip the code block.  Code blocks can
    # be skipped more than once.
    for block in code_blocks:
        for pattern in args.skips:
            if block.contents.find(pattern) > -1:
                block.skip(pattern)


def print_report(args: Args, blocks: List[FencedBlock]) -> None:
    """Print Markdown fenced block report and skips report."""
    report = []
    filename = click.format_filename(args.markdown_file)
    title1 = filename + ' fenced blocks'
    text1 = fenced_block_report(blocks, title=title1)
    report.append(text1)

    roles = [b.role.name for b in blocks]
    counts = Counter(roles)

    report.append('{} test cases'.format(counts['CODE']))
    if counts['SKIP_CODE'] > 0:
        report.append('{} skipped code blocks'.format(
            counts['SKIP_CODE']
        ))

    num_missing_output = counts['CODE'] - counts['OUTPUT']
    report.append(
        '{} code blocks missing an output block'.format(
            num_missing_output
        )
    )

    if args.skips:
        report.append('')
        title2 = 'skip pattern matches (blank means no match)'
        text2 = skips_report(args.skips, blocks, title=title2)
        report.append(text2)
    print('\n'.join(report))


def fenced_block_report(blocks: List[FencedBlock], title: str = '') -> str:
    """Generate text report about the input file fenced code blocks."""
    table = monotable.MonoTable()
    table.max_cell_height = 7
    table.more_marker = '...'
    cell_grid = []
    for block in blocks:
        if block.role == Role.SKIP_CODE:
            quoted_skips = [r.join(['"', '"']) for r in block.skip_reasons]
            skips = '\n'.join(quoted_skips)
        else:
            skips = ''
        cell_grid.append([block.type, block.line, block.role.value, skips])
    headings = [
        'block\ntype', 'line\nnumber', 'test\nrole',
        'skip pattern/reason\nquoted and one per line']
    formats = ['', '', '', '(width=30)']
    text = table.table(headings, formats, cell_grid, title)    # type: str
    return text


def skips_report(
        skips: List[str], blocks: List[FencedBlock], title: str = '') -> str:
    """Generate text report about the disposition of --skip options."""
    table = monotable.MonoTable()
    table.max_cell_height = 5
    table.more_marker = '...'
    cell_grid = []
    for skip in skips:
        code_lines = []
        for block in blocks:
            if skip in block.skip_reasons:
                code_lines.append(str(block.line))

        cell_grid.append([skip, ', '.join(code_lines)])
    headings = ['skip pattern', 'matching code block line number(s)']
    formats = ['', '(width=36;wrap)']
    text = table.table(headings, formats, cell_grid, title)    # type: str
    return text


def test_nothing_fails() -> None:
    """Fail if no Python code blocks were processed."""
    assert False, 'nothing to test'


def test_nothing_passes() -> None:
    """Succeed  if no Python code blocks were processed."""
    # nothing to test
    pass


_ASSERTION_MESSAGE = 'zero length {} block at line {}'


def build_test_cases(args: Args, blocks: List[FencedBlock]) -> str:
    """Generate test code from the Python fenced code blocks."""
    # repr escapes back slashes from win filesystem paths
    # so it can be part of the generated test module docstring.
    quoted_markdown_path = repr(click.format_filename(args.markdown_file))
    markdown_path = quoted_markdown_path[1:-1]
    docstring_text = 'pytest file built from {}'.format(markdown_path)
    builder = print_capture.PytestFile(docstring_text)
    number_of_test_cases = 0
    for block in blocks:
        if block.role == Role.CODE:
            code_identifier = 'code_' + str(block.line)
            output_identifier = ''
            code = block.contents
            assert code, _ASSERTION_MESSAGE.format(
                'code', block.line)

            output_block = block.get_output()
            if output_block:
                output_identifier = '_output_' + str(output_block.line)
                expected_output = output_block.contents
                assert expected_output, _ASSERTION_MESSAGE.format(
                    'expected output', block.line)
            else:
                expected_output = ''
            identifier = code_identifier + output_identifier
            builder.add_test_case(identifier, code, expected_output)
            number_of_test_cases += 1
    if number_of_test_cases == 0:
        if args.fail_nocode:
            test_function = inspect.getsource(test_nothing_fails)
        else:
            test_function = inspect.getsource(test_nothing_passes)
        builder.add_source(test_function)
    return str(builder)
